@@ -90,33 +90,22 @@
 
 #ifndef DOUBLE
 
+#define DEFAULT_THRESHOLD (256 * DOUBLE_EPS)
+
+
 #define DEFINE_COMMON
 #define DOUBLE double
-#define DOUBLE_MANT_DIG DBL_MANT_DIG
+#define DOUBLE_EPS DBL_EPSILON
 #define SUFFIX(_n) _n
 #define FORMAT_f "f"
 #define FORMAT_g "g"
 
 #ifdef GOFFICE_WITH_LONG_DOUBLE
 
-#if 0
-/* We define this to cause certain "double" functions to be implemented in
-   terms of their "long double" counterparts.  */
-#define BOUNCE
-static GORegressionResult
-general_linear_regressionl (long double *const *const xssT, int n,
-			    const long double *ys, int m,
-			    long double *result,
-			    go_regression_stat_tl *stat_,
-			    gboolean affine);
-#endif
-
-
 #include "go-regression.c"
-#undef BOUNCE
 #undef DEFINE_COMMON
 #undef DOUBLE
-#undef DOUBLE_MANT_DIG
+#undef DOUBLE_EPS
 #undef SUFFIX
 #undef FORMAT_f
 #undef FORMAT_g
@@ -124,7 +113,7 @@ general_linear_regressionl (long double *const *const xssT, int n,
 #include <sunmath.h>
 #endif
 #define DOUBLE long double
-#define DOUBLE_MANT_DIG LDBL_MANT_DIG
+#define DOUBLE_EPS LDBL_EPSILON
 #define SUFFIX(_n) _n ## l
 #define FORMAT_f "Lf"
 #define FORMAT_g "Lg"
@@ -221,6 +210,19 @@ go_regression_statl_get_type (void)
 	       (dst)[_i] = (src)[_i];		\
   } while (0)
 
+#define PRINT_MATRIX(var,dim1,dim2)				\
+  do {								\
+	int _i, _j, _d1, _d2;					\
+	_d1 = (dim1);						\
+	_d2 = (dim2);						\
+	for (_i = 0; _i < _d1; _i++) {				\
+	  for (_j = 0; _j < _d2; _j++) {			\
+		  g_printerr (" %12.8" FORMAT_g, (var)[_i][_j]); \
+	  }							\
+	  g_printerr ("\n");					\
+	}							\
+  } while (0)
+
 #define PRINT_QMATRIX(var,dim1,dim2)				\
   do {								\
 	int _i, _j, _d1, _d2;					\
@@ -251,303 +253,66 @@ go_regression_statl_get_type (void)
  *
  */
 
-/* ------------------------------------------------------------------------- */
-#ifdef BOUNCE
-/* Suppport routines for bouncing double to long double.  */
-
-static double *
-shrink_vector (const long double *V, int d)
+static SUFFIX(GOQuadMatrix)
+*
+SUFFIX(quad_matrix_from_matrix) (CONSTMATRIX A, int m, int n)
 {
-	if (V) {
-		double *V2 = g_new (double, d);
-		COPY_VECTOR (V2, V, d);
-		return V2;
-	} else
-		return NULL;
+	int i, j;
+	SUFFIX(GOQuadMatrix) *qA = SUFFIX(go_quad_matrix_new) (m, n);
+
+	for (i = 0; i < m; i++)
+		for (j = 0; j < n; j++)
+			SUFFIX(go_quad_init) (&qA->data[i][j], A[i][j]);
+
+	return qA;
 }
 
 static void
-copy_stats (go_regression_stat_t *s1,
-	    const go_regression_stat_tl *s2, int n)
+SUFFIX(copy_quad_matrix_to_matrix) (MATRIX A, const SUFFIX(GOQuadMatrix) *qA)
 {
-	if (!s2)
-		return;
+	int i, j;
 
-	g_free (s1->se);
-	g_free (s1->t);
-	g_free (s1->xbar);
+	for (i = 0; i < qA->m; i++)
+		for (j = 0; j < qA->n; j++)
+			A[i][j] = SUFFIX(go_quad_value) (&qA->data[i][j]);
 
-	s1->se = shrink_vector (s2->se, n);
-	s1->t = shrink_vector (s2->t, n);
-	s1->sqr_r = s2->sqr_r;
-	s1->adj_sqr_r = s2->adj_sqr_r;
-	s1->se_y = s2->se_y;
-	s1->F = s2->F;
-	s1->df_reg = s2->df_reg;
-	s1->df_resid = s2->df_resid;
-	s1->df_total = s2->df_total;
-	s1->ss_reg = s2->ss_reg;
-	s1->ss_resid = s2->ss_resid;
-	s1->ss_total = s2->ss_total;
-	s1->ms_reg = s2->ms_reg;
-	s1->ms_resid = s2->ms_resid;
-	s1->ybar = s2->ybar;
-	s1->xbar = shrink_vector (s2->xbar, n);
-	s1->var = s2->var;
 }
-#endif
 
 /* ------------------------------------------------------------------------- */
 
-/*
- * QR decomposition of a matrix using Householder matrices.
- *
- * A (input) is an m-times-n matrix.  A[0...m-1][0..n-1]
- * If qAT is TRUE, this parameter is transposed.
- *
- * V is a pre-allocated output m-times-n matrix.  V will contrain
- * n vectors of different lengths: n, n-1, ..., 1.  These are the
- * Householder vectors (or null for the degenerate case).
- *
- * Rfinal is a pre-allocated output matrix of size n-times-n.
- *
- * qdet is an optional output location for det(Q).
- *
- * Note: the output is V and R, not Q and R.
- */
-static gboolean
-SUFFIX(QRH) (CONSTMATRIX A, gboolean qAT,
-	     QMATRIX V, QMATRIX Rfinal, int m, int n,
-	     int *qdet)
-{
-	QMATRIX R = NULL;
-	int i, j, k;
-	QUAD *tmp = g_new (QUAD, n);
-	gboolean ok = TRUE;
-
-	if (m < n || n < 1) {
-		ok = FALSE;
-		goto out;
-	}
-
-	ALLOC_MATRIX2 (R, m, n, QUAD);
-	for (i = 0; i < m; i++) {
-		for (j = 0; j < n; j++) {
-			DOUBLE Aij = qAT ? A[j][i] : A[i][j];
-			if (!SUFFIX(go_finite)(Aij)) {
-				ok = FALSE;
-				goto out;
-			}
-			SUFFIX(go_quad_init) (&R[i][j], Aij);
-			SUFFIX(go_quad_init) (&V[i][j], -42);
-		}
-	}
-
-	if (qdet)
-		*qdet = 1;
-
-	for (k = 0; k < n; k++) {
-		QUAD L, L2, L2p, s;
-
-		SUFFIX(go_quad_init) (&L2, 0);
-		L2p = L2;
-		for (i = m - 1; i >= k; i--) {
-			V[i][k] = R[i][k];
-			SUFFIX(go_quad_mul)(&s, &V[i][k], &V[i][k]);
-			L2p = L2;
-			SUFFIX(go_quad_add)(&L2, &L2, &s);
-		}
-		SUFFIX(go_quad_sqrt)(&L, &L2);
-
-		(SUFFIX(go_quad_value)(&V[k][k]) < 0
-		 ? SUFFIX(go_quad_sub)
-		 : SUFFIX(go_quad_add)) (&V[k][k], &V[k][k], &L);
-
-		/* Normalize v[k] to length 1.  */
-		SUFFIX(go_quad_mul)(&s, &V[k][k], &V[k][k]);
-		SUFFIX(go_quad_add)(&L2p, &L2p, &s);
-		SUFFIX(go_quad_sqrt)(&L, &L2p);
-		if (SUFFIX(go_quad_value)(&L) == 0) {
-			/* This will be an identity so no determinant sign */
-			continue;
-		}
-		for (i = k; i < m; i++)
-			SUFFIX(go_quad_div)(&V[i][k], &V[i][k], &L);
-
-		/* Householder matrices have determinant -1.  */
-		if (qdet)
-			*qdet = -*qdet;
-
-		/* Calculate tmp = v[k]^t * R[k:m,k:n] */
-		for (j = k; j < n; j++) {
-			SUFFIX(go_quad_init) (&tmp[j], 0);
-			for (i = k ; i < m; i++) {
-				QUAD p;
-				SUFFIX(go_quad_mul) (&p, &V[i][k], &R[i][j]);
-				SUFFIX(go_quad_add) (&tmp[j], &tmp[j], &p);
-			}
-		}
-
-		/* R[k:m,k:n] -= v[k] * tmp */
-		for (j = k; j < n; j++) {
-			for (i = k; i < m; i++) {
-				QUAD p;
-				SUFFIX(go_quad_mul) (&p, &V[i][k], &tmp[j]);
-				SUFFIX(go_quad_add) (&p, &p, &p);
-				SUFFIX(go_quad_sub) (&R[i][j], &R[i][j], &p);
-			}
-		}
-
-		/* Explicitly zero what should become zero.  */
-		for (i = k + 1; i < m; i++)
-			SUFFIX(go_quad_init) (&R[i][k], 0);
-
 #if 0
-		g_printerr ("After round %d:\n", k);
-		g_printerr ("R:\n");
-		PRINT_QMATRIX(R, m ,n );
-		g_printerr ("\n");
-		g_printerr ("V:\n");
-		PRINT_QMATRIX(V, m ,n );
-		g_printerr ("\n\n");
-#endif
-	}
-
-	for (i = 0; i < n /* Only n */; i++)
-		for (j = 0; j < n; j++)
-			Rfinal[i][j] = R[i][j];
-
-out:
-	if (R)
-		FREE_MATRIX (R, m, n);
-	g_free (tmp);
-
-	return ok;
-}
-
 static void
-SUFFIX(multiply_Qt) (QUAD *x, CONSTQMATRIX V, int m, int n)
-{
-	int i, k;
-
-	for (k = 0; k < n; k++) {
-		QUAD s;
-		SUFFIX(go_quad_init)(&s, 0);
-		for (i = k; i < m; i++) {
-			QUAD p;
-			SUFFIX(go_quad_mul) (&p, &x[i], &V[i][k]);
-			SUFFIX(go_quad_add) (&s, &s, &p);
-		}
-		SUFFIX(go_quad_add) (&s, &s, &s);
-		for (i = k; i < m; i++) {
-			QUAD p;
-			SUFFIX(go_quad_mul) (&p, &s, &V[i][k]);
-			SUFFIX(go_quad_sub) (&x[i], &x[i], &p);
-		}
-	}
-}
-
-/*
- * Solve Rx=b.
- *
- * Return TRUE in case of error.
- */
-static gboolean
-SUFFIX(back_solve) (CONSTQMATRIX R, QUAD *x, const QUAD *b, int n)
+SUFFIX(print_Q) (CONSTQMATRIX V, int m, int n)
 {
+	QMATRIX Q;
 	int i, j;
+	QUAD *x;
 
-	for (i = n - 1; i >= 0; i--) {
-		QUAD d = b[i];
-		for (j = i + 1; j < n; j++) {
-			QUAD p;
-			SUFFIX(go_quad_mul) (&p, &R[i][j], &x[j]);
-			SUFFIX(go_quad_sub) (&d, &d, &p);
-		}
-		if (SUFFIX(go_quad_value)(&R[i][i]) == 0) {
-			while (i >= 0) SUFFIX(go_quad_init) (&x[i--], 0);
-			return TRUE;
-		}
-		SUFFIX(go_quad_div) (&x[i], &d, &R[i][i]);
+	ALLOC_MATRIX2 (Q, m, m, QUAD);
+	x = g_new (QUAD, m);
+
+	for (j = 0; j < m; j++) {
+		for (i = 0; i < m; i++)
+			SUFFIX(go_quad_init)(&x[i], i == j ? 1 : 0);
+
+		SUFFIX(multiply_Qt)(x, V, m, n);
+		for (i = 0; i < m; i++)
+			Q[j][i] = x[i];
 	}
 
-	return FALSE;
+	PRINT_QMATRIX (Q, m, m);
+
+	FREE_MATRIX (Q, m, m);
+	g_free (x);
 }
-
-/*
- * Solve RTx=b.
- *
- * Return TRUE in case of error.
- */
-static gboolean
-SUFFIX(fwd_solve) (CONSTQMATRIX R, QUAD *x, const QUAD *b, int n)
-{
-	int i, j;
-
-	for (i = 0; i < n; i++) {
-		QUAD d = b[i];
-		for (j = 0; j < i; j++) {
-			QUAD p;
-			SUFFIX(go_quad_mul) (&p, &R[j][i], &x[j]);
-			SUFFIX(go_quad_sub) (&d, &d, &p);
-		}
-		if (SUFFIX(go_quad_value)(&R[i][i]) == 0) {
-			while (i < n) SUFFIX(go_quad_init) (&x[i], 0);
-			return TRUE;
-		}
-		SUFFIX(go_quad_div) (&x[i], &d, &R[i][i]);
-	}
-
-	return FALSE;
-}
-
-static GORegressionResult
-SUFFIX(regres_from_condition) (CONSTQMATRIX R, int n)
-{
-	DOUBLE emin = go_pinf;
-	DOUBLE emax = go_ninf;
-	DOUBLE c, lc;
-	int i;
-
-	/*
-	 * R is triangular, so all the eigenvalues are in the diagonal.
-	 * We need the absolute largest and smallest.
-	 */
-	for (i = 0; i < n; i++) {
-		DOUBLE e = SUFFIX(fabs)(SUFFIX(go_quad_value)(&R[i][i]));
-		if (e < emin) emin = e;
-		if (e > emax) emax = e;
-	}
-
-	if (emin == 0)
-		return GO_REG_singular;
-
-	/* Condition number */
-	c = emax / emin;
-
-	/* Number of bits destroyed by matrix.  Since the calculations are
-	   done with QUADs, we can afford to lose a lot.  */
-	lc = SUFFIX(log)(c) / SUFFIX(log)(FLT_RADIX);
-
-#if 1
-	if (lc > 0.95* DOUBLE_MANT_DIG)
-		return GO_REG_near_singular_bad;
-	if (lc > 0.75 * DOUBLE_MANT_DIG)
-		return GO_REG_near_singular_good;
 #endif
 
-	return GO_REG_ok;
-}
-
-#ifndef BOUNCE
-
-static void
+static DOUBLE
 SUFFIX(calc_residual) (CONSTMATRIX AT, const DOUBLE *b, int m, int n,
-		       const QUAD *y, QUAD *residual, QUAD *N2)
+		       const DOUBLE *y)
 {
 	int i, j;
-
-	SUFFIX(go_quad_init) (N2, 0);
+	QUAD N2 = SUFFIX(go_quad_zero);
 
 	for (i = 0; i < m; i++) {
 		QUAD d;
@@ -555,94 +320,18 @@ SUFFIX(calc_residual) (CONSTMATRIX AT, const DOUBLE *b, int m, int n,
 		SUFFIX(go_quad_init) (&d, b[i]);
 
 		for (j = 0; j < n; j++) {
-			QUAD Aij, e;
-			SUFFIX(go_quad_init) (&Aij, AT[j][i]);
-			SUFFIX(go_quad_mul) (&e, &Aij, &y[j]);
+			QUAD e;
+			SUFFIX(go_quad_mul12) (&e, AT[j][i], y[j]);
 			SUFFIX(go_quad_sub) (&d, &d, &e);
 		}
 
-		if (residual)
-			residual[i] = d;
-
 		SUFFIX(go_quad_mul) (&d, &d, &d);
-		SUFFIX(go_quad_add) (N2, N2, &d);
-	}
-}
-
-static void
-SUFFIX(refine) (CONSTMATRIX AT, const DOUBLE *b, int m, int n,
-		CONSTQMATRIX V, CONSTQMATRIX R, QUAD *result)
-{
-	QUAD *residual = g_new (QUAD, m);
-	QUAD *newresidual = g_new (QUAD, m);
-	QUAD *QTres = g_new (QUAD, m);
-	QUAD *delta = g_new (QUAD, n);
-	QUAD *newresult = g_new (QUAD, n);
-	int pass;
-	QUAD best_N2;
-
-	SUFFIX(calc_residual) (AT, b, m, n, result, residual, &best_N2);
-#ifdef DEBUG_REFINEMENT
-	g_printerr ("-: Residual norm = %20.15" FORMAT_g "\n",
-		    SUFFIX(go_quad_value) (&best_N2));
-#endif
-
-	for (pass = 1; pass < 10; pass++) {
-		int i;
-		QUAD dres, N2;
-
-		/* Compute Q^T residual.  */
-		for (i = 0; i < m; i++)
-			QTres[i] = residual[i];
-		SUFFIX(multiply_Qt)(QTres, V, m, n);
-
-		/* Solve R delta = Q^T residual */
-		if (SUFFIX(back_solve) (R, delta, QTres, n))
-			break;
-
-		/* newresult = result + delta */
-		for (i = 0; i < n; i++) {
-			SUFFIX(go_quad_add) (&newresult[i],
-					     &delta[i],
-					     &result[i]);
-#ifdef DEBUG_REFINEMENT
-			g_printerr ("d[%2d] = %20.15" FORMAT_f
-				    "  %20.15" FORMAT_f
-				    "  %20.15" FORMAT_f "\n",
-				    i,
-				    SUFFIX(go_quad_value) (&delta[i]),
-				    SUFFIX(go_quad_value) (&result[i]),
-				    SUFFIX(go_quad_value) (&newresult[i]));
-#endif
-		}
-
-		SUFFIX(calc_residual) (AT, b, m, n, newresult, newresidual, &N2);
-		SUFFIX (go_quad_sub) (&dres, &N2, &best_N2);
-
-#ifdef DEBUG_REFINEMENT
-		g_printerr ("%d: Residual norm = %20.15" FORMAT_g
-			    " (%.5" FORMAT_g ")\n",
-			    pass,
-			    SUFFIX(go_quad_value) (&N2),
-			    SUFFIX(go_quad_value) (&dres));
-#endif
-		if (SUFFIX(go_quad_value) (&dres) >= 0)
-			break;
-
-		best_N2 = N2;
-		COPY_VECTOR (result, newresult, n);
-		COPY_VECTOR (residual, newresidual, m);
+		SUFFIX(go_quad_add) (&N2, &N2, &d);
 	}
 
-	g_free (residual);
-	g_free (newresidual);
-	g_free (QTres);
-	g_free (newresult);
-	g_free (delta);
+	return SUFFIX(go_quad_value) (&N2);
 }
 
-#endif
-
 /* ------------------------------------------------------------------------- */
 
 /*
@@ -653,11 +342,13 @@ SUFFIX(refine) (CONSTMATRIX AT, const DOUBLE *b, int m, int n,
 GORegressionResult
 SUFFIX(go_linear_solve_multiple) (CONSTMATRIX A, MATRIX B, int n, int bn)
 {
-	QMATRIX V;
-	QMATRIX R;
 	void *state;
 	GORegressionResult regres;
-	gboolean ok;
+	SUFFIX(GOQuadMatrix) *qA;
+	const SUFFIX(GOQuadMatrix) *R;
+	SUFFIX(GOQuadQR) *qr;
+	QUAD *QTb, *x;
+	int i, j;
 
 	if (n < 1 || bn < 1)
 		return GO_REG_not_enough_data;
@@ -676,35 +367,37 @@ SUFFIX(go_linear_solve_multiple) (CONSTMATRIX A, MATRIX B, int n, int bn)
 
 	state = SUFFIX(go_quad_start) ();
 
-	ALLOC_MATRIX2 (V, n, n, QUAD);
-	ALLOC_MATRIX2 (R, n, n, QUAD);
-	ok = SUFFIX(QRH)(A, FALSE, V, R, n, n, NULL);
-	if (ok) {
-		QUAD *QTb = g_new (QUAD, n);
-		QUAD *x = g_new (QUAD, n);
-		int i, j;
-
-		regres = SUFFIX(regres_from_condition)(R, n);
-
-		for (j = 0; j < bn; j++) {
-			/* Compute Q^T b.  */
-			for (i = 0; i < n; i++)
-				SUFFIX(go_quad_init)(&QTb[i], B[i][j]);
-			SUFFIX(multiply_Qt)(QTb, V, n, n);
-
-			/* Solve R x = Q^T b  */
-			if (SUFFIX(back_solve) (R, x, QTb, n))
-				regres = GO_REG_singular;
-
-			/* Round for output.  */
-			for (i = 0; i < n; i++)
-				B[i][j] = SUFFIX(go_quad_value) (&x[i]);
-		}
-
-		g_free (x);
-		g_free (QTb);
-	} else
+	qA = SUFFIX(quad_matrix_from_matrix) (A, n, n);
+	qr = SUFFIX(go_quad_qr_new) (qA);
+	if (!qr) {
 		regres = GO_REG_invalid_data;
+		goto done;
+	}
+	R = SUFFIX(go_quad_qr_r) (qr);
+
+	QTb = g_new (QUAD, n);
+	x = g_new (QUAD, n);
+
+	for (j = 0; j < bn; j++) {
+		/* Compute Q^T b.  */
+		for (i = 0; i < n; i++)
+			SUFFIX(go_quad_init)(&QTb[i], B[i][j]);
+		SUFFIX(go_quad_qr_multiply_qt)(qr, QTb);
+
+		/* Solve R x = Q^T b  */
+		if (SUFFIX(go_quad_matrix_back_solve) (R, x, QTb, FALSE))
+			regres = GO_REG_singular;
+
+		/* Round for output.  */
+		for (i = 0; i < n; i++)
+			B[i][j] = SUFFIX(go_quad_value) (&x[i]);
+	}
+
+	SUFFIX(go_quad_qr_free) (qr);
+	g_free (x);
+	g_free (QTb);
+done:
+	SUFFIX(go_quad_matrix_free) (qA);
 
 	SUFFIX(go_quad_end) (state);
 
@@ -739,144 +432,62 @@ SUFFIX(go_linear_solve) (CONSTMATRIX A, const DOUBLE *b, int n, DOUBLE *res)
 gboolean
 SUFFIX(go_matrix_invert) (MATRIX A, int n)
 {
-	QMATRIX V;
-	QMATRIX R;
-	gboolean has_result;
-	void *state = SUFFIX(go_quad_start) ();
+	SUFFIX(GOQuadMatrix) *qA, *qZ;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
+	void *state;
+	gboolean ok;
 
-	ALLOC_MATRIX2 (V, n, n, QUAD);
-	ALLOC_MATRIX2 (R, n, n, QUAD);
-
-	has_result = SUFFIX(QRH)(A, FALSE, V, R, n, n, NULL);
-
-	if (has_result) {
-		int i, k;
-		QUAD *x = g_new (QUAD, n);
-		QUAD *QTk = g_new (QUAD, n);
-		GORegressionResult regres =
-			SUFFIX(regres_from_condition) (R, n);
-
-		if (regres != GO_REG_ok && regres != GO_REG_near_singular_good)
-			has_result = FALSE;
-
-		for (k = 0; has_result && k < n; k++) {
-			/* Compute Q^T's k-th column.  */
-			for (i = 0; i < n; i++)
-				SUFFIX(go_quad_init)(&QTk[i], i == k);
-			SUFFIX(multiply_Qt)(QTk, V, n, n);
-
-			/* Solve R x = Q^T e_k */
-			if (SUFFIX(back_solve) (R, x, QTk, n))
-				has_result = FALSE;
-
-			/* Round for output.  */
-			for (i = 0; i < n; i++)
-				A[i][k] = SUFFIX(go_quad_value) (&x[i]);
-		}
-
-		g_free (QTk);
-		g_free (x);
+	state = SUFFIX(go_quad_start) ();
+	qA = SUFFIX(quad_matrix_from_matrix) (A, n, n);
+	qZ = SUFFIX(go_quad_matrix_inverse) (qA, threshold);
+	ok = (qZ != NULL);
+	SUFFIX(go_quad_matrix_free) (qA);
+	if (qZ) {
+		SUFFIX(copy_quad_matrix_to_matrix) (A, qZ);
+		SUFFIX(go_quad_matrix_free) (qZ);
 	}
-
-	FREE_MATRIX (V, n, n);
-	FREE_MATRIX (R, n, n);
-
 	SUFFIX(go_quad_end) (state);
-
-	return has_result;
+	return ok;
 }
 
 DOUBLE
 SUFFIX(go_matrix_determinant) (CONSTMATRIX A, int n)
 {
-	gboolean ok;
-	QMATRIX R;
-	QMATRIX V;
-	int qdet;
 	DOUBLE res;
+	QUAD qres;
 	void *state;
+	SUFFIX(GOQuadMatrix) *qA;
 
 	if (n < 1)
 		return 0;
 
-	/* Special case.  */
-	if (n == 1)
-		return A[0][0];
-
 	state = SUFFIX(go_quad_start) ();
 
-	/* Special case.  */
-	if (n == 2) {
-		QUAD a, b, det;
-		SUFFIX(go_quad_mul12)(&a, A[0][0], A[1][1]);
-		SUFFIX(go_quad_mul12)(&b, A[1][0], A[0][1]);
-		SUFFIX(go_quad_sub)(&det, &a, &b);
-		res = SUFFIX(go_quad_value)(&det);
-		goto done;
-	}
+	qA = SUFFIX(quad_matrix_from_matrix) (A, n, n);
+	SUFFIX(go_quad_matrix_determinant) (qA, &qres);
+	SUFFIX(go_quad_matrix_free) (qA);
+	res = SUFFIX(go_quad_value) (&qres);
 
-	ALLOC_MATRIX2 (V, n, n, QUAD);
-	ALLOC_MATRIX2 (R, n, n, QUAD);
-	ok = SUFFIX(QRH)(A, FALSE, V, R, n, n, &qdet);
-	if (ok) {
-		int i;
-		QUAD det;
-		SUFFIX(go_quad_init)(&det, qdet);
-		for (i = 0; i < n; i++)
-			SUFFIX(go_quad_mul)(&det, &det, &R[i][i]);
-		res = SUFFIX(go_quad_value)(&det);
-	} else
-		res = go_nan;
-
-	FREE_MATRIX (V, n, n);
-	FREE_MATRIX (R, n, n);
-
-done:
 	SUFFIX(go_quad_end) (state);
 
 	return res;
 }
 
 /* ------------------------------------------------------------------------- */
-/* Note, that this function takes a transposed matrix xssT.  */
 
+/* Note, that this function takes a transposed matrix xssT.  */
 static GORegressionResult
 SUFFIX(general_linear_regression) (CONSTMATRIX xssT, int n,
 				   const DOUBLE *ys, int m,
 				   DOUBLE *result,
 				   SUFFIX(go_regression_stat_t) *stat_,
-				   gboolean affine)
+				   gboolean affine,
+				   DOUBLE threshold)
 {
 	GORegressionResult regerr;
-#ifdef BOUNCE
-	long double **xssT2, *ys2, *result2;
-	go_regression_stat_tl *stat2;
-
-	ALLOC_MATRIX2 (xssT2, n, m, long double);
-	COPY_MATRIX (xssT2, xssT, n, m);
-
-	ys2 = g_new (long double, m);
-	COPY_VECTOR (ys2, ys, m);
-
-	result2 = g_new (long double, n);
-	stat2 = stat_ ? go_regression_stat_newl () : NULL;
-
-	regerr = general_linear_regressionl (xssT2, n, ys2, m,
-					     result2, stat2, affine);
-	COPY_VECTOR (result, result2, n);
-	copy_stats (stat_, stat2, n);
-
-	go_regression_stat_destroyl (stat2);
-	g_free (result2);
-	g_free (ys2);
-	FREE_MATRIX (xssT2, n, m);
-#else
-	QMATRIX V;
-	QMATRIX R;
-	QUAD *qresult;
-	QUAD *QTy;
-	QUAD *inv;
-	int i, k;
+	SUFFIX(GOQuadMatrix) *xss;
+	SUFFIX(GOQuadQR) *qr;
+	int i, j, k;
 	gboolean has_result;
 	void *state;
 	gboolean has_stat;
@@ -896,34 +507,52 @@ SUFFIX(general_linear_regression) (CONSTMATRIX xssT, int n,
 
 	state = SUFFIX(go_quad_start) ();
 
-	ALLOC_MATRIX2 (V, m, n, QUAD);
-	ALLOC_MATRIX2 (R, n, n, QUAD);
-	qresult = g_new0 (QUAD, n);
-	QTy = g_new (QUAD, m);
-	inv = g_new (QUAD, n);
+	xss = SUFFIX(go_quad_matrix_new) (m, n);
+	for (i = 0; i < m; i++)
+		for (j = 0; j < n; j++)
+			SUFFIX(go_quad_init) (&xss->data[i][j], xssT[j][i]);
+	qr = SUFFIX(go_quad_qr_new) (xss);
+	SUFFIX(go_quad_matrix_free) (xss);
 
-	has_result = SUFFIX(QRH) (xssT, TRUE, V, R, m, n, NULL);
+	has_result = (qr != NULL);
 
 	if (has_result) {
+		const SUFFIX(GOQuadMatrix) *R = SUFFIX(go_quad_qr_r) (qr);
+		QUAD *qresult = g_new0 (QUAD, n);
+		QUAD *QTy = g_new (QUAD, m);
+		QUAD *inv = g_new (QUAD, n);
+		DOUBLE emax;
+		int df_resid = m - n;
+		int df_reg = n - (affine ? 1 : 0);
+
 		regerr = GO_REG_ok;
 
+		SUFFIX(go_quad_matrix_eigen_range) (R, NULL, &emax);
+		for (i = 0; i < n; i++) {
+			DOUBLE ei = SUFFIX(go_quad_value)(&R->data[i][i]);
+			gboolean degenerate =
+				!(SUFFIX(fabs)(ei) >= emax * threshold);
+			if (degenerate) {
+				SUFFIX(go_quad_qr_mark_degenerate) (qr, i);
+				df_resid++;
+				df_reg--;
+			}
+		}
+
 		/* Compute Q^T ys.  */
 		for (i = 0; i < m; i++)
 			SUFFIX(go_quad_init)(&QTy[i], ys[i]);
-		SUFFIX(multiply_Qt)(QTy, V, m, n);
+		SUFFIX(go_quad_qr_multiply_qt)(qr, QTy);
+
+		if (0)
+			SUFFIX(go_quad_matrix_dump) (R, "%10.5" FORMAT_g);
 
 		/* Solve R res = Q^T ys */
-		if (SUFFIX(back_solve) (R, qresult, QTy, n))
+		if (SUFFIX(go_quad_matrix_back_solve) (R, qresult, QTy, TRUE))
 			has_result = FALSE;
 
-		if (n > 2)
-			SUFFIX(refine) (xssT, ys, m, n, V, R, qresult);
-
-		/* Round to plain precision.  */
-		for (i = 0; i < n; i++) {
+		for (i = 0; i < n; i++)
 			result[i] = SUFFIX(go_quad_value) (&qresult[i]);
-			SUFFIX(go_quad_init) (&qresult[i], result[i]);
-		}
 
 		/* This should not fail since n >= 1.  */
 		err = SUFFIX(go_range_average) (ys, m, &stat_->ybar);
@@ -943,8 +572,9 @@ SUFFIX(general_linear_regression) (CONSTMATRIX xssT, int n,
 			g_assert (err == 0);
 		}
 
-		SUFFIX(calc_residual) (xssT, ys, m, n, qresult, NULL, &N2);
-		stat_->ss_resid = SUFFIX(go_quad_value) (&N2);
+		;
+		stat_->ss_resid =
+			SUFFIX(calc_residual) (xssT, ys, m, n, result);
 
 		stat_->sqr_r = (stat_->ss_total == 0)
 			? 1
@@ -959,49 +589,32 @@ SUFFIX(general_linear_regression) (CONSTMATRIX xssT, int n,
 
 		/* FIXME: we want to guard against division by zero.  */
 		stat_->adj_sqr_r = 1 - stat_->ss_resid * (m - 1) /
-			((m - n) * stat_->ss_total);
-		if (m == n)
-			SUFFIX(go_quad_init) (&N2, 0);
+			(df_resid * stat_->ss_total);
+		if (df_resid == 0)
+			N2 = SUFFIX(go_quad_zero);
 		else {
 			QUAD d;
-			SUFFIX(go_quad_init) (&d, m - n);
+			SUFFIX(go_quad_init) (&d, df_resid);
 			SUFFIX(go_quad_div) (&N2, &N2, &d);
 		}
 		stat_->var = SUFFIX(go_quad_value) (&N2);
 
 		stat_->se = g_new0 (DOUBLE, n);
 		for (k = 0; k < n; k++) {
-			QUAD p;
+			QUAD p, N;
 
 			/* inv = e_k */
 			for (i = 0; i < n; i++)
 				SUFFIX(go_quad_init) (&inv[i], i == k ? 1 : 0);
 
 			/* Solve R^T inv = e_k */
-			if (SUFFIX(fwd_solve) (R, inv, inv, n)) {
+			if (SUFFIX(go_quad_matrix_fwd_solve) (R, inv, inv, TRUE)) {
 				regerr = GO_REG_singular;
 				break;
 			}
 
-			/* Solve R newinv = inv */
-			if (SUFFIX(back_solve) (R, inv, inv, n)) {
-				regerr = GO_REG_singular;
-				break;
-			}
-
-			if (SUFFIX(go_quad_value) (&inv[k]) < 0) {
-				/*
-				 * If this happens, something is really
-				 * wrong, numerically.
-				 */
-				g_printerr ("inv[%d]=%" FORMAT_g "\n",
-					    k,
-					    SUFFIX(go_quad_value) (&inv[k]));
-				regerr = GO_REG_near_singular_bad;
-				break;
-			}
-
-			SUFFIX(go_quad_mul) (&p, &N2, &inv[k]);
+			SUFFIX(go_quad_dot_product) (&N, inv, inv, n);
+			SUFFIX(go_quad_mul) (&p, &N2, &N);
 			SUFFIX(go_quad_sqrt) (&p, &p);
 			stat_->se[k] = SUFFIX(go_quad_value) (&p);
 		}
@@ -1013,37 +626,36 @@ SUFFIX(general_linear_regression) (CONSTMATRIX xssT, int n,
 				? SUFFIX(go_pinf)
 				: result[i] / stat_->se[i];
 
-		stat_->df_resid = m - n;
-		stat_->df_reg = n - (affine ? 1 : 0);
-		stat_->df_total = stat_->df_resid + stat_->df_reg;
+		stat_->df_resid = df_resid;
+		stat_->df_reg = df_reg;
+		stat_->df_total = stat_->df_resid + df_reg;
 
 		stat_->F = (stat_->sqr_r == 1)
 			? SUFFIX(go_pinf)
-			: ((stat_->sqr_r / stat_->df_reg) /
+			: ((stat_->sqr_r / df_reg) /
 			   (1 - stat_->sqr_r) * stat_->df_resid);
 
 		stat_->ss_reg =  stat_->ss_total - stat_->ss_resid;
 		stat_->se_y = SUFFIX(sqrt) (stat_->ss_total / m);
-		stat_->ms_reg = (stat_->df_reg == 0)
+		stat_->ms_reg = (df_reg == 0)
 			? 0
 			: stat_->ss_reg / stat_->df_reg;
-		stat_->ms_resid = (stat_->df_resid == 0)
+		stat_->ms_resid = (df_resid == 0)
 			? 0
-			: stat_->ss_resid / stat_->df_resid;
+			: stat_->ss_resid / df_resid;
+
+		g_free (QTy);
+		g_free (qresult);
+		g_free (inv);
 	} else
 		regerr = GO_REG_invalid_data;
 
 	SUFFIX(go_quad_end) (state);
 
-	g_free (QTy);
-	g_free (qresult);
-	g_free (inv);
-	FREE_MATRIX (V, m, n);
-	FREE_MATRIX (R, n, n);
+	if (qr) SUFFIX(go_quad_qr_free) (qr);
 out:
 	if (!has_stat)
 		SUFFIX(go_regression_stat_destroy) (stat_);
-#endif
 
 	return regerr;
 }
@@ -1261,6 +873,7 @@ SUFFIX(go_linear_regression) (MATRIX xss, int dim,
 			      SUFFIX(go_regression_stat_t) *stat_)
 {
 	GORegressionResult result;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
 
 	g_return_val_if_fail (dim >= 1, GO_REG_invalid_dimensions);
 	g_return_val_if_fail (n >= 1, GO_REG_invalid_dimensions);
@@ -1276,14 +889,14 @@ SUFFIX(go_linear_regression) (MATRIX xss, int dim,
 
 		result = SUFFIX(general_linear_regression)
 			(xss2, dim + 1, ys, n,
-			 res, stat_, affine);
+			 res, stat_, affine, threshold);
 		g_free (xss2[0]);
 		g_free (xss2);
 	} else {
 		res[0] = 0;
 		result = SUFFIX(general_linear_regression)
 			(xss, dim, ys, n,
-			 res + 1, stat_, affine);
+			 res + 1, stat_, affine, threshold);
 	}
 	return result;
 }
@@ -1349,6 +962,7 @@ SUFFIX(go_exponential_regression_as_log) (MATRIX xss, int dim,
 	DOUBLE *log_ys;
 	GORegressionResult result;
 	int i;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
 
 	g_return_val_if_fail (dim >= 1, GO_REG_invalid_dimensions);
 	g_return_val_if_fail (n >= 1, GO_REG_invalid_dimensions);
@@ -1373,14 +987,14 @@ SUFFIX(go_exponential_regression_as_log) (MATRIX xss, int dim,
 
 		result = SUFFIX(general_linear_regression)
 			(xss2, dim + 1, log_ys,
-			 n, res, stat_, affine);
+			 n, res, stat_, affine, threshold);
 		g_free (xss2[0]);
 		g_free (xss2);
 	} else {
 		res[0] = 0;
 		result = SUFFIX(general_linear_regression)
 			(xss, dim, log_ys, n,
-			 res + 1, stat_, affine);
+			 res + 1, stat_, affine, threshold);
 	}
 
  out:
@@ -1414,6 +1028,7 @@ SUFFIX(go_power_regression) (MATRIX xss, int dim,
 	DOUBLE *log_ys, **log_xss = NULL;
 	GORegressionResult result;
 	int i, j;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
 
 	g_return_val_if_fail (dim >= 1, GO_REG_invalid_dimensions);
 	g_return_val_if_fail (n >= 1, GO_REG_invalid_dimensions);
@@ -1448,14 +1063,14 @@ SUFFIX(go_power_regression) (MATRIX xss, int dim,
 
 		result = SUFFIX(general_linear_regression)
 			(log_xss2, dim + 1, log_ys,
-			 n, res, stat_, affine);
+			 n, res, stat_, affine, threshold);
 		g_free (log_xss2[0]);
 		g_free (log_xss2);
 	} else {
 		res[0] = 0;
 		result = SUFFIX(general_linear_regression)
 			(log_xss, dim, log_ys, n,
-			 res + 1, stat_, affine);
+			 res + 1, stat_, affine, threshold);
 	}
 
  out:
@@ -1496,6 +1111,7 @@ SUFFIX(go_logarithmic_regression) (MATRIX xss, int dim,
         MATRIX log_xss;
 	GORegressionResult result;
 	int i, j;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
 
 	g_return_val_if_fail (dim >= 1, GO_REG_invalid_dimensions);
 	g_return_val_if_fail (n >= 1, GO_REG_invalid_dimensions);
@@ -1522,14 +1138,14 @@ SUFFIX(go_logarithmic_regression) (MATRIX xss, int dim,
 
 		result = SUFFIX(general_linear_regression)
 			(log_xss2, dim + 1, ys, n,
-			 res, stat_, affine);
+			 res, stat_, affine, threshold);
 		g_free (log_xss2[0]);
 		g_free (log_xss2);
 	} else {
 		res[0] = 0;
 		result = SUFFIX(general_linear_regression)
 			(log_xss, dim, ys, n,
-			 res + 1, stat_, affine);
+			 res + 1, stat_, affine, threshold);
 	}
 
  out:
@@ -2074,49 +1690,49 @@ SUFFIX(go_non_linear_regression) (SUFFIX(GORegressionFunction) f,
 GORegressionResult
 SUFFIX(go_linear_regression_leverage) (MATRIX A, DOUBLE *d, int m, int n)
 {
-	QMATRIX V;
-	QMATRIX R;
-	gboolean has_result;
+	SUFFIX(GOQuadMatrix) *qA;
+	SUFFIX(GOQuadQR) *qr;
 	void *state = SUFFIX(go_quad_start) ();
 	GORegressionResult regres;
+	DOUBLE threshold = DEFAULT_THRESHOLD;
 
-	ALLOC_MATRIX2 (V, m, n, QUAD);
-	ALLOC_MATRIX2 (R, n, n, QUAD);
-
-	has_result = SUFFIX(QRH)(A, FALSE, V, R, m, n, NULL);
-
-	if (has_result) {
+	qA = SUFFIX(quad_matrix_from_matrix) (A, m, n);
+	qr = SUFFIX(go_quad_qr_new) (qA);
+	if (qr) {
 		int k;
 		QUAD *b = g_new (QUAD, n);
+		const SUFFIX(GOQuadMatrix) *R = SUFFIX(go_quad_qr_r) (qr);
+		DOUBLE emin, emax;
 
-		regres = SUFFIX(regres_from_condition) (R, n);
+		SUFFIX(go_quad_matrix_eigen_range) (R, &emin, &emax);
+		regres = (emin > emax * threshold)
+			? GO_REG_ok
+			: GO_REG_singular;
 
 		for (k = 0; k < m; k++) {
 			int i;
-			QUAD acc;
+			QUAD acc = SUFFIX(go_quad_zero);
 
 			/* b = AT e_k  */
 			for (i = 0; i < n; i++)
-				SUFFIX(go_quad_init) (&b[i], A[k][i]);
+				b[i] = qA->data[k][i];
 
-			/* Solve R^T newb = b */
-			if (SUFFIX(fwd_solve) (R, b, b, n)) {
+			/* Solve R^T b = AT e_k */
+			if (SUFFIX(go_quad_matrix_fwd_solve) (R, b, b, FALSE)) {
 				regres = GO_REG_singular;
 				break;
 			}
 
 			/* Solve R newb = b */
-			if (SUFFIX(back_solve) (R, b, b, n)) {
+			if (SUFFIX(go_quad_matrix_back_solve) (R, b, b, FALSE)) {
 				regres = GO_REG_singular;
 				break;
 			}
 
 			/* acc = (Ab)_k */
-			SUFFIX(go_quad_init) (&acc, 0);
 			for (i = 0; i < n; i++) {
 				QUAD p;
-				SUFFIX(go_quad_init) (&p, A[k][i]);
-				SUFFIX(go_quad_mul) (&p, &p, &b[i]);
+				SUFFIX(go_quad_mul) (&p, &qA->data[k][i], &b[i]);
 				SUFFIX(go_quad_add) (&acc, &acc, &p);
 			}
 
@@ -2124,13 +1740,47 @@ SUFFIX(go_linear_regression_leverage) (MATRIX A, DOUBLE *d, int m, int n)
 		}
 
 		g_free (b);
+		SUFFIX(go_quad_qr_free) (qr);
 	} else
 		regres = GO_REG_invalid_data;
 
-	FREE_MATRIX (V, m, n);
-	FREE_MATRIX (R, n, n);
-
 	SUFFIX(go_quad_end) (state);
 
 	return regres;
 }
+
+/*
+ * Compute the pseudo-inverse of a matrix, see
+ *     http://en.wikipedia.org/wiki/Moore%E2%80%93Penrose_pseudoinverse
+ *
+ * Think of this as A+ = (AT A)^-1 AT, except that the pseudo-inverse
+ * is defined even if AT*A isn't invertible.
+ *
+ * Properties:
+ *
+ *     A  A+ A  = A
+ *     A+ A  A+ = A+
+ *     (A A+)^T = A A+
+ *     (A+ A)^T = A+ A
+ *
+ * assuming threshold==0 and no rounding errors.  
+ */
+
+void
+SUFFIX(go_matrix_pseudo_inverse) (CONSTMATRIX A, int m, int n,
+				  DOUBLE threshold,
+				  MATRIX B)
+{
+	SUFFIX(GOQuadMatrix) *qA, *qZ;
+	void *state;
+
+	state = SUFFIX(go_quad_start) ();
+	qA = SUFFIX(quad_matrix_from_matrix) (A, m, n);
+	qZ = SUFFIX(go_quad_matrix_pseudo_inverse) (qA, threshold);
+	SUFFIX(go_quad_matrix_free) (qA);
+	if (qZ) {
+		SUFFIX(copy_quad_matrix_to_matrix) (B, qZ);
+		SUFFIX(go_quad_matrix_free) (qZ);
+	}
+	SUFFIX(go_quad_end) (state);
+}
